# -*- coding: utf-8 -*-
import unittest
from datetime import datetime

from exactitude import dates


class DatesTest(unittest.TestCase):

    def test_is_partial_date(self):
        self.assertTrue(dates.validate('2017-04-04 10:30:29'))
        self.assertTrue(dates.validate('2017-04-04 10:30'))
        self.assertTrue(dates.validate('2017-04-04 10'))
        self.assertTrue(dates.validate('2017-04-04'))
        self.assertTrue(dates.validate('2017-4-4'))
        self.assertTrue(dates.validate('2017-4'))
        self.assertTrue(dates.validate('2017'))
        self.assertFalse(dates.validate('0017'))
        self.assertFalse(dates.validate(None))
        self.assertFalse(dates.validate(5))
        self.assertFalse(dates.validate('2017-20-01'))

    def test_chop_dates(self):
        self.assertEquals(dates.clean('2017-00-00'), '2017')
        self.assertEquals(dates.clean('2017-00-00T00:00:00'), '2017')
        self.assertEquals(dates.clean('2017-00-00T12:03:49'), '2017')
        self.assertEquals(dates.clean('2017-01-01T00:00:00'), '2017-01-01')

    def test_convert_datetime(self):
        dt = datetime.utcnow()
        iso, _ = dt.isoformat().split('.', 1)
        self.assertEquals(dates.clean(dt), iso)
        self.assertTrue(dates.validate(iso))

        dt = datetime.utcnow().date()
        iso = dt.isoformat()
        self.assertEquals(dates.clean(dt), iso)

    def test_parse_date(self):
        self.assertEquals(dates.clean(None), None)
        self.assertEquals(dates.clean(''), None)
        self.assertEquals(dates.clean('2017-04-04'), '2017-04-04')
        # self.assertEquals(parse_date('2017-4-4'), '2017-04-04')

        # TODO: make this yield an imprecise date somehow?
        self.assertEquals(dates.clean('4/2017', format="%m/%Y"),
                          '2017-04-01')
        self.assertEquals(dates.clean('4/2xx017', format="%m/%Y"),
                          None)

    def test_guess_date(self):
        self.assertEquals(dates.clean('12.4.2017'), '2017-04-12')

    def test_fuzzy_date_parser_failure(self):
        self.assertEquals(dates.fuzzy_date_parser('nothing'), None)

    def test_fuzzy_date_parser_success_english(self):
        result = dates.fuzzy_date_parser('15 march, 1987')

        self.assertIsInstance(result, datetime)
        self.assertEqual(result.strftime('%x'), '03/15/87')

    def test_fuzzy_date_parser_success_german(self):
        result = dates.fuzzy_date_parser(u'15. März 1987')

        self.assertIsInstance(result, datetime)
        self.assertEqual(result.strftime('%x'), '03/15/87')

    def test_fuzzy_date_parser_success_spanish(self):
        result = dates.fuzzy_date_parser(u'15 Marzo 1987')

        self.assertIsInstance(result, datetime)
        self.assertEqual(result.strftime('%x'), '03/15/87')



import countrynames
from normality import stringify

from exactitude.common import ExactitudeType


class CountryType(ExactitudeType):

    def __init__(self, *args):
        super(CountryType, self).__init__(*args)
        # extra countries that OCCRP is interested in.
        self.names = {
            'zz': 'Global',
            'eu': 'European Union',
            'xk': 'Kosovo',
            'yucs': 'Yugoslavia',
            'csxx': 'Serbia and Montenegro',
            'suhh': 'Soviet Union',
            'ge-ab': 'Abkhazia',
            'x-so': 'South Ossetia',
            'so-som': 'Somaliland',
            'gb-wls': 'Wales',
            'gb-sct': 'Scotland',
            'md-pmr': 'Transnistria'
        }
        for code, label in self.locale.territories.items():
            self.names[code.lower()] = label

    def validate(self, country, **kwargs):
        country = stringify(country)
        if country is None:
            return False
        return country.lower() in self.names

    def clean_text(self, country, guess=False, **kwargs):
        """Determine a two-letter country code based on an input.

        The input may be a country code, a country name, etc.
        """
        code = country.lower().strip()
        if code in self.names:
            return code
        country = countrynames.to_code(country, fuzzy=guess)
        if country is not None:
            return country.lower()



from exactitude.url import UrlType
from exactitude.name import NameType
from exactitude.domain import DomainType
from exactitude.email import EmailType
from exactitude.address import AddressType
from exactitude.date import DateType
from exactitude.phone import PhoneType
from exactitude.country import CountryType
from exactitude.language import LanguageType
from exactitude.identifier import IdentifierType
from exactitude.common import TextType

urls = UrlType()
names = NameType()
domains = DomainType()
emails = EmailType()
addresses = AddressType()
dates = DateType()
phones = PhoneType()
countries = CountryType()
languages = LanguageType()
identifiers = IdentifierType()
texts = TextType()

__all__ = [urls,
           names,
           domains,
           emails,
           addresses,
           dates,
           phones,
           countries,
           languages,
           identifiers,
           texts]



import unittest

from exactitude import identifiers


class IdentifiersTest(unittest.TestCase):

    def test_parse(self):
        self.assertEqual(identifiers.clean('88/9'), '88/9')

    def test_normalize(self):
        self.assertEqual(identifiers.normalize('FOO'), ['foo'])
        self.assertEqual(identifiers.normalize('xx '), ['xx'])
        self.assertEqual(identifiers.normalize(' '), [])

    def test_domain_validity(self):
        self.assertTrue(identifiers.validate('foo@pudo.org'))


from __future__ import unicode_literals

import re
from normality import stringify

from exactitude.common import ExactitudeType
from exactitude.domain import DomainType


class EmailType(ExactitudeType):
    EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")
    domains = DomainType()

    def validate(self, email, **kwargs):
        """Check to see if this is a valid email address."""
        email = stringify(email)
        if email is None:
            return
        if not self.EMAIL_REGEX.match(email):
            return False
        mailbox, domain = email.rsplit('@', 1)
        return self.domains.validate(domain, **kwargs)

    def clean_text(self, email, **kwargs):
        """Parse and normalize an email address.

        Returns None if this is not an email address.
        """
        if not self.EMAIL_REGEX.match(email):
            return None
        mailbox, domain = email.rsplit('@', 1)
        domain = self.domains.clean(domain, **kwargs)
        if domain is None or mailbox is None:
            return
        return '@'.join((mailbox, domain))

    def normalize(self, email, **kwargs):
        """Normalize for comparison."""
        emails = super(EmailType, self).normalize(email, **kwargs)
        return [e.lower() for e in emails]



import unittest

from exactitude import emails


class EmailsTest(unittest.TestCase):

    def test_parse(self):
        self.assertEqual(emails.clean('foo@pudo.org'), 'foo@pudo.org')
        self.assertEqual(emails.clean('pudo.org'), None)
        self.assertEqual(emails.clean(None), None)
        self.assertEqual(emails.clean(5), None)
        self.assertEqual(emails.clean('foo@PUDO.org'), 'foo@pudo.org')
        self.assertEqual(emails.clean('FOO@PUDO.org'), 'FOO@pudo.org')

    def test_normalize(self):
        self.assertEqual(emails.normalize('FOO@PUDO'), [])
        self.assertEqual(emails.normalize('FOO@PUDO.org'), ['foo@pudo.org'])

    def test_domain_validity(self):
        self.assertTrue(emails.validate('foo@pudo.org'))
        self.assertFalse(emails.validate('foo@pudo'))
        self.assertFalse(emails.validate('@pudo.org'))
        self.assertFalse(emails.validate('foo@'))



import re
from normality import normalize as normalize_text
from normality.cleaning import collapse_spaces

from exactitude.common import ExactitudeType


class AddressType(ExactitudeType):
    LINE_BREAKS = re.compile(r'(\r\n|\n|<BR/>|<BR>|\t|ESQ\.,|ESQ,|;)')
    COMMATA = re.compile(r'(,\s?[,\.])')

    def clean_text(self, address, **kwargs):
        """Basic clean-up."""
        address = self.LINE_BREAKS.sub(', ', address)
        address = self.COMMATA.sub(', ', address)
        address = collapse_spaces(address)
        if len(address):
            return address

    def normalize(self, address, **kwargs):
        """Make the address more compareable."""
        # TODO: normalize well-known parts like "Street", "Road", etc.
        # TODO: consider using https://github.com/openvenues/pypostal
        addresses = super(AddressType, self).normalize(address, **kwargs)
        return [normalize_text(a) for a in addresses]



import unittest

from exactitude import domains


class DomainsTest(unittest.TestCase):

    def test_parse(self):
        self.assertEqual(domains.clean('pudo.org'), 'pudo.org')
        self.assertEqual(domains.clean('pudoorg'), None)
        self.assertEqual(domains.clean(None), None)
        self.assertEqual(domains.clean('x.a'), None)

    def test_normalize(self):
        self.assertEqual(domains.normalize('PUDO'), [])
        self.assertEqual(domains.normalize('PUDO.org'), ['pudo.org'])

    def test_domain_validity(self):
        self.assertTrue(domains.validate('pudo.org'))
        self.assertFalse(domains.validate('pudo'))
        self.assertFalse(domains.validate('x.a'))
        self.assertFalse(domains.validate(''))
        self.assertFalse(domains.validate('@pudo.org'))



import unittest

from exactitude import countries


class CountriesTest(unittest.TestCase):

    def test_country_codes(self):
        self.assertEqual(countries.clean('DE'), 'de')
        self.assertTrue(countries.validate('DE'))
        self.assertFalse(countries.validate('DEU'))
        self.assertFalse(countries.validate('SU'))
        self.assertTrue(countries.validate('XK'))
        self.assertTrue(countries.validate('EU'))

    def test_country_names(self):
        self.assertEqual(countries.clean(None), None)
        self.assertEqual(countries.clean('Takatukaland', guess=False), None)
        self.assertEqual(countries.clean('Germany'), 'de')
        # self.assertEqual(countries.clean('Germani'), 'de')
        self.assertEqual(countries.clean('Soviet Union'), 'suhh')



# import socket
from normality import stringify
from six.moves.urllib.parse import urlparse

from exactitude.common import ExactitudeType


class DomainType(ExactitudeType):
    # TODO: https://pypi.python.org/pypi/publicsuffix/

    # def _check_exists(self, domain):
    #     """Actually try to resolve a domain name."""
    #     try:
    #         domain = domain.encode('idna').lower()
    #         socket.getaddrinfo(domain, None)
    #         return True
    #     except:
    #         return False

    def validate(self, obj, **kwargs):
        """Check if a thing is a valid domain name."""
        text = stringify(obj)
        if text is None:
            return False
        if '.' not in text:
            return False
        if '@' in text or ':' in text:
            return False
        if len(text) < 4:
            return False
        return True

    def clean_text(self, domain, **kwargs):
        """Try to extract only the domain bit from the """
        try:
            # handle URLs by extracting the domain name
            domain = urlparse(domain).hostname or domain
            domain = domain.lower()
            # get rid of port specs
            domain = domain.rsplit(':', 1)[0]
            domain = domain.rstrip('.')
            # handle unicode
            domain = domain.encode("idna").decode('ascii')
        except ValueError:
            return None
        if self.validate(domain):
            return domain



from babel import Locale
from banal import ensure_list
from normality import stringify


class ExactitudeType(object):
    """Base class for all types."""

    def __init__(self, locale='en_GB'):
        self.locale = Locale(locale)

    def validate(self, text, **kwargs):
        """Returns a boolean to indicate if this is a valid instance of
        the type."""
        cleaned = self.clean(text, **kwargs)
        return cleaned is not None

    def clean(self, text, **kwargs):
        """Create a more clean, but still user-facing version of an
        instance of the type."""
        text = stringify(text)
        if text is not None:
            return self.clean_text(text, **kwargs)

    def clean_text(self, text, **kwargs):
        return text

    def normalize(self, text, cleaned=False, **kwargs):
        """Create a represenation ideal for comparisons, but not to be
        shown to the user."""
        if not cleaned:
            text = self.clean(text, **kwargs)
        return ensure_list(text)


class TextType(ExactitudeType):
    pass



from normality.cleaning import collapse_spaces

from exactitude.common import ExactitudeType


class NameType(ExactitudeType):

    def clean_text(self, name, **kwargs):
        """Basic clean-up."""
        return collapse_spaces(name)



import unittest

from exactitude import addresses

UK = """43 Duke Street
Edinburgh
EH6 8HH"""


class AddressesTest(unittest.TestCase):

    def test_clean(self):
        self.assertEqual(addresses.clean(UK),
                         '43 Duke Street, Edinburgh, EH6 8HH')
        self.assertEqual(addresses.clean('huhu\n   haha'), 'huhu, haha')
        self.assertEqual(addresses.clean('huhu,\n haha'), 'huhu, haha')

    def test_normalize(self):
        self.assertEqual(addresses.normalize(UK), ['43 duke street edinburgh eh6 8hh'])  # noqa
        self.assertEqual(addresses.normalize('\n  '), [])



import unittest

from exactitude import languages


class LanguagesTest(unittest.TestCase):

    def test_validate(self):
        self.assertTrue(languages.validate('de'))
        self.assertTrue(languages.validate('en'))
        self.assertFalse(languages.validate('us'))
        self.assertFalse(languages.validate(None))

    def test_cleam(self):
        self.assertEquals(languages.clean('de'), 'de')
        self.assertEquals(languages.clean('xx'), None)
        self.assertEquals(languages.clean(None), None)



from normality import stringify

from exactitude.common import ExactitudeType


class LanguageType(ExactitudeType):

    def __init__(self, *args):
        super(LanguageType, self).__init__(*args)
        self.names = {}
        for code, label in self.locale.languages.items():
            self.names[code.lower()] = label

    def validate(self, text, **kwargs):
        text = stringify(text)
        if text is None:
            return False
        return text.lower() in self.names

    def clean_text(self, text, **kwargs):
        code = text.lower().strip()
        if code in self.names:
            return code



import unittest

from exactitude import names


class NamesTest(unittest.TestCase):

    def test_parse(self):
        self.assertEqual(names.clean('Hans Well'), 'Hans Well')
        self.assertEqual(names.clean('Hans   Well '), 'Hans Well')

    def test_normalize(self):
        self.assertEqual(names.normalize('FOO'), ['FOO'])
        self.assertEqual(names.normalize('xx '), ['xx'])
        self.assertEqual(names.normalize(' '), [])

    def test_domain_validity(self):
        self.assertTrue(names.validate('huhu'))
        self.assertFalse(names.validate(''))



import unittest

from exactitude import urls


class UrlsTest(unittest.TestCase):

    def test_is_url(self):
        self.assertTrue(urls.validate('http://foo.org'))
        self.assertFalse(urls.validate(None))
        self.assertFalse(urls.validate('hello'))

    def test_parse_url(self):
        self.assertEqual(urls.clean('http://foo.com'), 'http://foo.com/')
        self.assertEqual(urls.clean('http://foo.com/#lala'), 'http://foo.com/')

        self.assertEqual(urls.clean('http://foo.com?b=1&a=2'),
                         'http://foo.com/?a=2&b=1')
        self.assertEqual(urls.clean('http://FOO.com'), 'http://foo.com/')
        self.assertEqual(urls.clean('http://FOO.com/A'), 'http://foo.com/A')



from setuptools import setup, find_packages


setup(
    name='exactitude',
    version='2.1.0',
    description="A library with real-world data parsers.",
    long_description="",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.6'
    ],
    keywords='names countries phones domains email country',
    author='Friedrich Lindenberg',
    author_email='friedrich@pudo.org',
    url='http://github.com/alephdata/dalet',
    license='MIT',
    packages=find_packages(exclude=['ez_setup', 'examples', 'test']),
    namespace_packages=[],
    package_data={},
    include_package_data=True,
    zip_safe=False,
    test_suite='nose.collector',
    install_requires=[
        'six >= 1.11.0',
        'normality >= 0.5.3',
        'babel >= 2.5.1',
        'urlnormalizer >= 1.0.3',
        'countrynames >= 1.2',
        'parsedatetime >= 2.1',
        'phonenumbers >= 8.8.4'
    ],
    tests_require=['nose'],
    entry_points={}
)



from urlnormalizer import normalize_url, is_valid_url

from exactitude.common import ExactitudeType


class UrlType(ExactitudeType):

    def validate(self, url, **kwargs):
        """Check if `url` is a valid URL."""
        return is_valid_url(url)

    def clean_text(self, url, **kwargs):
        """Perform intensive care on URLs, see `urlnormalizer`."""
        return normalize_url(url)



import re
import parsedatetime
from normality import stringify
from datetime import datetime, date

from exactitude.common import ExactitudeType


class DateType(ExactitudeType):
    # JS: '^([12]\\d{3}(-[01]?[1-9](-[0123]?[1-9])?)?)?$'
    DATE_RE = re.compile('^([12]\d{3}(-[01]?[0-9](-[0123]?[0-9]([T ]([012]?\d(:\d{1,2}(:\d{1,2})?)?)?)?)?)?)?$')  # noqa
    CUT_ZEROES = re.compile(r'((\-00.*)|(.00:00:00))$')
    MAX_LENGTH = 19

    def validate(self, obj, **kwargs):
        """Check if a thing is a valid date."""
        obj = stringify(obj)
        if obj is None:
            return False
        return self.DATE_RE.match(obj) is not None

    def clean_datetime(self, obj):
        """Python objects want to be text."""
        if isinstance(obj, datetime):
            return obj.isoformat()[:self.MAX_LENGTH]
        if isinstance(obj, date):
            return obj.isoformat()

    def clean(self, text, guess=True, format=None, **kwargs):
        """The classic: date parsing, every which way."""
        # handle date/datetime before converting to text.
        date = self.clean_datetime(text)
        if date is not None:
            return date

        text = stringify(text)
        if text is None:
            return

        if format is not None:
            # parse with a specified format
            try:
                obj = datetime.strptime(text, format)
                return obj.date().isoformat()
            except:
                return None

        if guess and not self.validate(text):
            # use dateparser to guess the format
            obj = self.fuzzy_date_parser(text)
            if obj is not None:
                return obj.date().isoformat()

        # limit to the date part of a presumed date string
        text = text[:self.MAX_LENGTH]
        # strip -00-00 from dates because it makes ES barf.
        text = self.CUT_ZEROES.sub('', text)
        text = text.replace(' ', 'T')

        if self.validate(text):
            return text

    def fuzzy_date_parser(self, text):
        """Thin wrapper around ``parsedatetime`` module.
        Since there's no upstream suppport for multiple locales, this wrapper
        exists.
        :param str text: Text to parse.
        :returns: A parsed date/time object. Raises exception on failure.
        :rtype: datetime
        """
        locales = parsedatetime._locales[:]

        # Loop through all the locales and try to parse successfully our string
        for locale in locales:
            const = parsedatetime.Constants(locale)
            const.re_option += re.UNICODE
            parser = parsedatetime.Calendar(const)
            parsed, ok = parser.parse(text)
            if ok:
                return datetime(*parsed[:6])



import unittest

from exactitude import phones


class PhonesTest(unittest.TestCase):

    def test_us_number(self):
        self.assertEqual(phones.clean('+1-800-784-2433'), '+18007842433')
        self.assertEqual(phones.clean('+1 800 784 2433'), '+18007842433')
        self.assertEqual(phones.clean('+18007842433'), '+18007842433')
        self.assertEqual(phones.clean('+1 555 8379'), None)

    def test_de_number(self):
        self.assertEqual(phones.clean('017623423980'), None)
        self.assertEqual(phones.clean('017623423980', countries='DE'),
                         '+4917623423980')



from normality import normalize

from exactitude.common import ExactitudeType


class IdentifierType(ExactitudeType):
    """Used for registration numbers, codes etc."""

    def normalize(self, text, **kwargs):
        """Normalize for comparison."""
        identifiers = []
        for ident in super(IdentifierType, self).normalize(text, **kwargs):
            identifiers.append(normalize(ident))
        return identifiers



import six
from banal import ensure_list
from phonenumbers import parse as parse_number
from phonenumbers import is_possible_number, is_valid_number, format_number
from phonenumbers import PhoneNumberFormat
from phonenumbers.phonenumberutil import NumberParseException

from exactitude.common import ExactitudeType


class PhoneType(ExactitudeType):

    def _clean_countries(self, countries, country):
        result = set([None])
        countries = ensure_list(countries)
        countries.extend(ensure_list(country))
        for country in countries:
            if isinstance(country, six.string_types):
                country = country.strip().upper()
                result.add(country)
        return result

    def clean_text(self, number, countries=None, country=None, **kwargs):
        """Parse a phone number and return in international format.

        If no valid phone number can be detected, None is returned. If
        a country code is supplied, this will be used to infer the
        prefix.

        https://github.com/daviddrysdale/python-phonenumbers
        """
        for code in self._clean_countries(countries, country):
            try:
                num = parse_number(number, code)
                if is_possible_number(num):
                    if is_valid_number(num):
                        return format_number(num, PhoneNumberFormat.E164)
            except NumberParseException:
                pass
